@@ -78,10 +78,6 @@ Place, Suite 330, Boston, MA 02111-1307 USA
 #include "xtrabackup.h"
 #include "xtrabackup_config.h"
 #include "xtrabackup_version.h"
-#ifdef HAVE_VERSION_CHECK
-#include <version_check_pl.h>
-#endif
-
 
 /** Possible values for system variable "innodb_checksum_algorithm". */
 extern const char *innodb_checksum_algorithm_names[];
@@ -428,8 +424,8 @@ bool backup_file_printf(const char *filename, const char *fmt, ...) {
 }
 
 template <typename F>
-static bool run_data_threads(const char *dir, F func, uint n,
-                             const char *thread_description) {
+bool run_data_threads(const char *dir, const char *suffix, F func, uint n,
+                      const char *thread_description) {
   datadir_thread_ctxt_t *data_threads;
   uint i, count;
   ib_mutex_t count_mutex;
@@ -453,7 +449,7 @@ static bool run_data_threads(const char *dir, F func, uint n,
   }
 
   xb_process_datadir(
-      dir, "",
+      dir, suffix,
       [&](const datadir_entry_t &entry, void *arg) mutable -> bool {
         queue.push(entry);
         return true;
@@ -488,6 +484,18 @@ static bool run_data_threads(const char *dir, F func, uint n,
   return (ret);
 }
 
+// Explicit instantiation for function pointer type:
+// void(*)(datadir_thread_ctxt_t*)
+template bool run_data_threads<void (*)(datadir_thread_ctxt_t *)>(
+    const char *, const char *, void (*)(datadir_thread_ctxt_t *), uint,
+    const char *);
+
+// Explicit instantiation for std::function<void(datadir_thread_ctxt_t*)> (used
+// when std::bind is applied)
+template bool run_data_threads<std::function<void(datadir_thread_ctxt_t *)>>(
+    const char *, const char *, std::function<void(datadir_thread_ctxt_t *)>,
+    uint, const char *);
+
 /************************************************************************
 Write buffer into .ibd file and preserve it's sparsiness. */
 bool write_ibd_buffer(ds_file_t *file, unsigned char *buf, size_t buf_len,
@@ -915,7 +923,7 @@ static bool copy_or_move_file(const char *src_file_path,
   ds_ctxt_t *datasink = ds_data; /* copy to datadir by default */
   bool ret;
 
-  /* File is located outsude of the datadir */
+  /* File is located outside of the datadir */
   char external_dir[FN_REFLEN];
 
   if (Fil_path::type_of_path(dst_file_path) == Fil_path::absolute) {
@@ -1044,7 +1052,7 @@ bool backup_files(const char *from, bool prep_mode, Backup_context &context) {
   xb::info() << "Starting " << (prep_mode ? "prep copy of" : "to backup")
              << " non-InnoDB tables and files";
 
-  run_data_threads(from,
+  run_data_threads(from, "" /* no suffix filtering */,
                    std::bind(backup_thread_func, std::placeholders::_1,
                              prep_mode, rsync_tmpfile, context),
                    xtrabackup_parallel, "backup");
@@ -1290,10 +1298,10 @@ Myrocks_checkpoint::file_list Myrocks_checkpoint::data_files() const {
   return Myrocks_datadir(checkpoint_dir).data_files();
 }
 
-static void par_copy_rocksdb_files(const Myrocks_datadir::const_iterator &start,
-                                   const Myrocks_datadir::const_iterator &end,
-                                   size_t thread_n, ds_ctxt_t *ds,
-                                   bool *result) {
+static void par_copy_or_move_rocksdb_files(
+    const Myrocks_datadir::const_iterator &start,
+    const Myrocks_datadir::const_iterator &end, size_t thread_n, ds_ctxt_t *ds,
+    bool *result) {
   for (auto it = start; it != end; it++) {
     if (ends_with(it->path.c_str(), ".qp") ||
         ends_with(it->path.c_str(), ".lz4") ||
@@ -1301,9 +1309,16 @@ static void par_copy_rocksdb_files(const Myrocks_datadir::const_iterator &start,
         ends_with(it->path.c_str(), ".xbcrypt")) {
       continue;
     }
-    if (!copy_file(ds, it->path.c_str(), it->rel_path.c_str(), thread_n,
-                   FILE_PURPOSE_OTHER, it->file_size)) {
-      *result = false;
+    if (xtrabackup_copy_back) {
+      if (!copy_file(ds, it->path.c_str(), it->rel_path.c_str(), thread_n,
+                     FILE_PURPOSE_OTHER, it->file_size)) {
+        *result = false;
+      }
+    } else {
+      if (!move_file(ds, it->path.c_str(), it->rel_path.c_str(), ds->root,
+                     thread_n, FILE_PURPOSE_OTHER)) {
+        *result = false;
+      }
     }
     if (!*result) {
       break;
@@ -1398,6 +1413,17 @@ bool backup_start(Backup_context &context) {
     /* LTFB/LIFB has to be executed before copying MyISAM */
     if (ddl_tracker != nullptr) {
       debug_sync_point("ddl_tracker_before_lock_ddl");
+
+      /* The tablespaces will be closed on handle_ddl_operations. Hence
+      dump the tablespace keys now. For tablespace that are tracked as
+      dropped, dont dump the tablespace encryption keys. Note that we
+      still need to save tablespace keys found from redo. We cannot do it
+      now as the redo log thread is still in progress. It is done after
+      the redo thread is stopped. See TablespaceKeyDumper::dump_from_redo() */
+      if (context.ts_key_dumper != nullptr) {
+        context.ts_key_dumper->dump_from_spaces(true);
+      }
+
       if (!lock_tables_for_backup(mysql_connection, opt_backup_lock_timeout,
                                   opt_backup_lock_retry_count)) {
         return (false);
@@ -1801,6 +1827,7 @@ bool copy_incremental_over_full() {
       xtrabackup::components::XTRABACKUP_KEYRING_FILE_CONFIG,
       xtrabackup::components::XTRABACKUP_KEYRING_KMIP_CONFIG,
       xtrabackup::components::XTRABACKUP_KEYRING_KMS_CONFIG,
+      xtrabackup::components::XTRABACKUP_KEYRING_VAULT_CONFIG,
       "ib_lru_dump",
       nullptr};
   bool ret = true;
@@ -1924,7 +1951,8 @@ bool copy_incremental_over_full() {
     bool result = true;
     std::function<void(const Myrocks_datadir::const_iterator &,
                        const Myrocks_datadir::const_iterator &, size_t)>
-        copy = std::bind(&par_copy_rocksdb_files, _1, _2, _3, ds_data, &result);
+        copy = std::bind(&par_copy_or_move_rocksdb_files, _1, _2, _3, ds_data,
+                         &result);
 
     par_for(PFS_NOT_INSTRUMENTED, rocksdb.files(), xtrabackup_parallel, copy);
   }
@@ -2010,6 +2038,7 @@ bool should_skip_file_on_copy_back(const char *filepath) {
       xtrabackup::components::XTRABACKUP_KEYRING_FILE_CONFIG,
       xtrabackup::components::XTRABACKUP_KEYRING_KMIP_CONFIG,
       xtrabackup::components::XTRABACKUP_KEYRING_KMS_CONFIG,
+      xtrabackup::components::XTRABACKUP_KEYRING_VAULT_CONFIG,
       ".qp",
       ".lz4",
       ".zst",
@@ -2374,8 +2403,9 @@ bool copy_back(int argc, char **argv) {
                << " threads for parallel data files transfer";
   }
 
-  ret = run_data_threads(".", copy_back_thread_func, xtrabackup_parallel,
-                         "copy-back");
+  ret =
+      run_data_threads(".", "" /* no suffix filtering */, copy_back_thread_func,
+                       xtrabackup_parallel, "copy-back");
   if (!ret) goto cleanup;
 
   /* copy buffer pool dump */
@@ -2423,7 +2453,8 @@ bool copy_back(int argc, char **argv) {
     using std::placeholders::_3;
     std::function<void(const Myrocks_datadir::const_iterator &,
                        const Myrocks_datadir::const_iterator &, size_t)>
-        copy = std::bind(&par_copy_rocksdb_files, _1, _2, _3, ds_data, &ret);
+        copy = std::bind(&par_copy_or_move_rocksdb_files, _1, _2, _3, ds_data,
+                         &ret);
 
     if (rocksdb_wal_dir.empty()) {
       par_for(PFS_NOT_INSTRUMENTED, rocksdb.files("", ""), xtrabackup_parallel,
@@ -2457,7 +2488,8 @@ bool copy_back(int argc, char **argv) {
       using std::placeholders::_3;
       std::function<void(const Myrocks_datadir::const_iterator &,
                          const Myrocks_datadir::const_iterator &, size_t)>
-          copy = std::bind(&par_copy_rocksdb_files, _1, _2, _3, ds_data, &ret);
+          copy = std::bind(&par_copy_or_move_rocksdb_files, _1, _2, _3, ds_data,
+                           &ret);
 
       par_for(PFS_NOT_INSTRUMENTED, rocksdb.wal_files(""), xtrabackup_parallel,
               copy);
@@ -2586,7 +2618,8 @@ bool decrypt_decompress_file(const char *filepath, uint thread_n) {
   }
   if (ds_data->fs_support_punch_hole) {
     char error[512];
-    if (!restore_sparseness(dest_filepath, opt_read_buffer_size, error)) {
+    if (!restore_sparseness(dest_filepath, opt_read_buffer_size,
+                            error IF_DEBUG(, true))) {
       xb::warn() << "restore_sparseness failed for file: " << dest_filepath
                  << " Error: " << error;
     }
@@ -2652,8 +2685,9 @@ bool decrypt_decompress() {
     xb::error() << "Error compiling filename regex";
     return (false);
   }
-  ret = run_data_threads(".", decrypt_decompress_thread_func,
-                         xtrabackup_parallel, "decrypt and decompress");
+  ret = run_data_threads(".", "" /* no suffix filtering */,
+                         decrypt_decompress_thread_func, xtrabackup_parallel,
+                         "decrypt and decompress");
 
   debug_sync_point("decrypt_decompress_func");
 
@@ -2670,39 +2704,35 @@ bool decrypt_decompress() {
   return (ret);
 }
 
-#ifdef HAVE_VERSION_CHECK
-void version_check() {
-  if (system("which perl > /dev/null 2>&1")) {
-    xb::info() << "perl binary not found. Skipping the version check";
-    return;
+void xtrabackup_apply_deltas_parallel(datadir_thread_ctxt_t *ctx) {
+  bool ret = true;
+  datadir_entry_t entry;
+  THD *thd = nullptr;
+
+  if (my_thread_init()) {
+    ret = false;
+    goto cleanup;
   }
 
-  if (opt_password != NULL) {
-    setenv("option_mysql_password", opt_password, 1);
-  }
-  if (opt_user != NULL) {
-    setenv("option_mysql_user", opt_user, 1);
-  }
-  if (opt_host != NULL) {
-    setenv("option_mysql_host", opt_host, 1);
-  }
-  if (opt_socket != NULL) {
-    setenv("option_mysql_socket", opt_socket, 1);
-  }
-  if (opt_port != 0) {
-    char port[20];
-    snprintf(port, sizeof(port), "%u", opt_port);
-    setenv("option_mysql_port", port, 1);
-  }
-  setenv("XTRABACKUP_VERSION", XTRABACKUP_VERSION, 1);
+  /* create THD to get thread number in the error log */
+  thd = create_thd(false, false, true, 0, 0);
 
-  FILE *pipe = popen("perl", "w");
-  if (pipe == NULL) {
-    return;
+  while (ctx->queue->pop(entry)) {
+    if (entry.is_empty_dir) continue;
+
+    ret = xtrabackup_apply_delta(entry, nullptr);
+    if (!ret) {
+      goto cleanup;
+    }
   }
 
-  fwrite((const char *)version_check_pl, version_check_pl_len, 1, pipe);
+cleanup:
+  my_thread_end();
+  destroy_thd(thd);
 
-  pclose(pipe);
+  mutex_enter(ctx->count_mutex);
+  --(*ctx->count);
+  mutex_exit(ctx->count_mutex);
+
+  ctx->ret = ret;
 }
-#endif
